@@ -3,5 +3,6 @@ from toolformer_pytorch.palm import PaLM
 from toolformer_pytorch.toolformer_pytorch import (
     Toolformer,
     filter_tokens_with_api_response,
-    sample
+    sample,
+    sample_with_api_call
 )
