@@ -1,4 +1,4 @@
-from functools import partial
+from functools import partial, wraps
 from collections import namedtuple
 
 import torch
@@ -47,9 +47,23 @@ def all_contains_id(t: torch.Tensor, token_id: int):
     mask = t == token_id
     return mask.any(dim = -1).all()
 
+def find_indices_of(t: torch.Tensor, token_id: int, occurrence = 1):
+    assert occurrence > 0
+    mask = (t == token_id)
+
+    has_occurred = mask.cumsum(dim = -1)
+    has_occurred = F.pad(has_occurred, (1, 0), value = 0.)
+
+    return (has_occurred < occurrence).sum(dim = -1).long()
+
+# invoking api call functions
+
+
+
 # sampling api related functions
 # they do greedy sampling, but encourage sampling api calls by auto-selecting <api> when that token is in the top k = 10
 
+@beartype
 @torch.no_grad()
 def sample(
     model: nn.Module,
@@ -81,7 +95,7 @@ def sample(
     # sampling positions - different sequences have different cursors
 
     positions = default(positions, torch.zeros((batch_size,), device = device, dtype = torch.long))
-    assert (positions <= prime_length).all() and (positions < max_seq_len).all(), 'all positions must be less then initial prime length as well as the total sequence length + 1 (plus one for noop if one sequence finished sampling before the other)'
+    assert (positions <= (prime_length + 1)).all() and (positions <= max_seq_len).all(), 'all positions must be less then initial prime length as well as the total sequence length + 1 (plus one for noop if one sequence finished sampling before the other)'
 
     # eval model
 
@@ -96,18 +110,59 @@ def sample(
     batch_indices = rearrange(batch_indices, 'b -> b 1')
     position_indices = rearrange(positions, 'b -> b 1')
 
+    # determine the <api> token mask, for making sure api is called only once, masking out logit to prevent it from being selected for those rows which already contains an <api> token
+
+    api_token_mask = None # lazily created, since do not know logit dimensions
+
+    def create_api_token_mask(num_tokens, api_start_token_id):
+        mask = torch.zeros((1, 1, num_tokens), dtype = torch.bool)
+        assert api_start_token_id < num_tokens
+        mask[..., api_start_token_id] = True
+        return mask
+
     # start iterating
 
     for iteration in tqdm(range(remain_iterations)):
         logits = model(output)
         last_logits = logits[batch_indices, position_indices]
 
+        if not exists(api_token_mask):
+            num_tokens = last_logits.shape[-1]
+            api_token_mask = create_api_token_mask(num_tokens, api_start_token_id)
+            api_token_mask = api_token_mask.to(device)
+
+        api_called = (output == api_start_token_id).any(dim = -1)
+
+        # this will ensure that each batch token sequence will have at most one <api> token
+
+        logit_mask = api_token_mask & rearrange(api_called, 'b -> b 1 1')
+        last_logits = last_logits.masked_fill(logit_mask, -torch.finfo(last_logits.dtype).max)
+
+        # greedy sample (but could be made non-greedy)
+
         sampled = gumbel_sample(last_logits, temperature = temperature)
 
+        # for those sequences without an api call, if the api_start_token_id is within top k (set to 10 in paper) of logits, just auto-select
+
+        # seems to be an important hack in the paper
+        # it seems like this paper will take a lot more follow up research to be viable
+
+        top_token_ids = last_logits.topk(select_api_start_id_top_k, dim = -1).indices
+        has_api_token_in_topk = (top_token_ids == api_start_token_id).any(dim = -1)
+        should_auto_select_api_token = has_api_token_in_topk & ~rearrange(api_called, 'b -> b 1')
+
+        sampled = sampled.masked_fill(should_auto_select_api_token, api_start_token_id)
+
+        # set the sampled tokens at the right curosr positions
+
         output[batch_indices, position_indices] = sampled
 
+        # increment positions
+
         position_indices += 1
-        position_indices.clamp_(max = seq_len)
+        position_indices.clamp_(max = seq_len) # noop if one sequence is further along and near the end
+
+        # if using <eos> tokens, look for all sequences having it and terminate, also anything after <eos> will be padded
 
         if exists(eos_token_id):
             eos_mask = (output == eos_token_id)
@@ -125,6 +180,46 @@ def sample(
 
     return output
 
+@beartype
+@torch.no_grad()
+def sample_with_api_call(
+    model: nn.Module,
+    *,
+    seq_len,
+    call_apis: Callable,
+    prime: torch.Tensor,
+    api_end_token_id: int,
+    occurrence = 1,
+    **kwargs
+):
+    sampled = sample(
+        model = model,
+        prime = prime,
+        seq_len = seq_len,
+        **kwargs
+    )
+
+    sampled = call_apis(sampled)
+
+    sampled_seq_len = sampled.shape[-1]
+    null_positions = sampled_seq_len # handle sequences that do not have api calls
+
+    pos_starting_at_end_of_api = find_indices_of(
+        sampled,
+        api_end_token_id,
+        occurrence = occurrence
+    )
+
+    resample_after_api_calls = sample(
+        model = model,
+        prime = sampled,
+        seq_len = sampled_seq_len,
+        positions = (pos_starting_at_end_of_api + 1).clamp(max = null_positions), # start at the position right after the </api>
+        **kwargs
+    )
+
+    return resample_after_api_calls
+
 # the main contribution of the paper is simply the filtering equations presented in section 2
 
 def default_weight_fn(t):
@@ -264,12 +359,14 @@ class Toolformer(nn.Module):
         self,
         model: nn.Module,
         tool: Callable,
-        teach_tool_prompt: str
+        teach_tool_prompt: str,
+        filters: dict[str, Callable[[str], bool]] = dict()
     ):
         super().__init__()
         self.model = model
         self.tool = tool
         self.teach_tool_prompt = teach_tool_prompt
+        self.filters = filters
 
     def forward(self):
         raise NotImplementedError
